
# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


from ._transform_ops_gen import _Dialect

@_ods_cext.register_operation(_Dialect)
class ApplyForLoopCanonicalizationPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns for canonicalizing operations inside SCF loop bodies.
  At the moment, only affine.min/max computations with iteration variables,
  loop bounds and loop steps are canonicalized.
  """

  OPERATION_NAME = "transform.apply_patterns.scf.for_loop_canonicalization"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(ApplyForLoopCanonicalizationPatternsOp)
class ApplyForLoopCanonicalizationPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_patterns.scf.for_loop_canonicalization"

def apply_patterns_scf_for_loop_canonicalization(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyForLoopCanonicalizationPatternsOp:
  return ApplyForLoopCanonicalizationPatternsOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplySCFStructuralConversionPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns for performing structural conversions of SCF operations.
  """

  OPERATION_NAME = "transform.apply_conversion_patterns.scf.structural_conversions"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(ApplySCFStructuralConversionPatternsOp)
class ApplySCFStructuralConversionPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_conversion_patterns.scf.structural_conversions"

def apply_conversion_patterns_scf_structural_conversions(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplySCFStructuralConversionPatternsOp:
  return ApplySCFStructuralConversionPatternsOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ApplySCFToControlFlowPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns that lower structured control flow ops to unstructured
  control flow.
  """

  OPERATION_NAME = "transform.apply_conversion_patterns.scf.scf_to_control_flow"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(ApplySCFToControlFlowPatternsOp)
class ApplySCFToControlFlowPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_conversion_patterns.scf.scf_to_control_flow"

def apply_conversion_patterns_scf_scf_to_control_flow(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplySCFToControlFlowPatternsOp:
  return ApplySCFToControlFlowPatternsOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ForallToForOp(_ods_ir.OpView):
  r"""
  Converts the `scf.forall` operation pointed to by the given handle into a
  set of nested `scf.for` operations. Each new operation corresponds to one
  induction variable of the original "multifor" loop.
  
  The operand handle must be associated with exactly one payload operation.
  
  Loops with shared outputs are currently not supported.
  
  #### Return Modes
  
  Consumes the operand handle. Produces a silenceable failure if the operand
  is not associated with a single `scf.forall` payload operation.
  Returns as many handles as the given `forall` op has induction variables
  that are associated with the generated `scf.for` loops.
  Produces a silenceable failure if another number of resulting handles is
  requested.
  """

  OPERATION_NAME = "transform.loop.forall_to_for"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _Sequence[_ods_ir.Type], target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.extend(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def transformed(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(ForallToForOp)
class ForallToForOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.forall_to_for"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def loop_forall_to_for(transformed: _Sequence[_ods_ir.Type], target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ForallToForOp]:
  op = ForallToForOp(transformed=transformed, target=target, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class ForallToParallelOp(_ods_ir.OpView):
  r"""
  Converts the `scf.forall` operation pointed to by the given handle into an
  `scf.parallel` operation.
  
  The operand handle must be associated with exactly one payload operation.
  
  Loops with outputs are not supported.
  
  #### Return Modes
  
  Consumes the operand handle. Produces a silenceable failure if the operand
  is not associated with a single `scf.forall` payload operation.
  Returns a handle to the new `scf.parallel` operation.
  Produces a silenceable failure if another number of resulting handles is
  requested.
  """

  OPERATION_NAME = "transform.loop.forall_to_parallel"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _Sequence[_ods_ir.Type], target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.extend(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def transformed(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(ForallToParallelOp)
class ForallToParallelOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.forall_to_parallel"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def loop_forall_to_parallel(transformed: _Sequence[_ods_ir.Type], target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ForallToParallelOp]:
  op = ForallToParallelOp(transformed=transformed, target=target, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class HoistLoopInvariantSubsetsOp(_ods_ir.OpView):
  r"""
  This transform hoists loop-invariant subset ops out of the targeted
  loop-like op. It looks for matching subset extraction/insertion op pairs and
  hoists them. The loop body operates on a newly introduced region iter_arg.
  
  Subset ops are hoisted only from the targeted op. If subset ops should be
  hoisted from an entire loop nest, this transformation must be applied to
  each loop-like op of the loop nest, starting with the innermost loop and
  ending with the outermost loop.
  
  Example:
  ```
  %r = scf.for ... iter_args(%t = %a) -> (tensor<?xf32>) {
    %0 = tensor.extract_slice %t[0][5][1] : tensor<?xf32> to tensor<5xf32>
    %1 = "test.foo"(%0) : (tensor<5xf32>) -> (tensor<5xf32>)
    %2 = tensor.insert_slice %1 into %t[0][5][1]
        : tensor<5xf32> into tensor<?xf32>
    scf.yield %2 : tensor<?xf32>
  }
  ```
  Is transformed to:
  ```
  %0 = tensor.extract_slice %a[0][5][1] : tensor<?xf32> to tensor<5xf32>
  %new_loop:2 = scf.for ... iter_args(%t = %a, %h = %0) -> (tensor<?xf32>) {
    %1 = "test.foo"(%h) : (tensor<5xf32>) -> (tensor<5xf32>)
    scf.yield %t, %2 : tensor<?xf32>, tensor<5xf32>
  }
  %r = tensor.insert_slice %new_loop#1 into %new_loop#0
      : tensor<5xf32> into tensor<?xf32>
  ```
  
  Subset ops are hoisted only if there are no conflicting subset ops. E.g.,
  if there were a second overlapping extraction in the above example, no ops
  could be hoisted safely.
  
  This transform reads the target handle and modifies the payload. This
  transform does not invalidate any handles, but loop-like ops are replaced
  with new loop-like ops when a subset op is hoisted. The transform rewriter
  updates all handles accordingly.
  """

  OPERATION_NAME = "transform.loop.hoist_loop_invariant_subsets"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(HoistLoopInvariantSubsetsOp)
class HoistLoopInvariantSubsetsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.hoist_loop_invariant_subsets"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def loop_hoist_loop_invariant_subsets(target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> HoistLoopInvariantSubsetsOp:
  return HoistLoopInvariantSubsetsOp(target=target, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class LoopCoalesceNestedOp(_ods_ir.OpView):
  r"""
  Given a loop, collects all nested loops (including imperfectly nested
  loops with operations between them) and coalesces them directly using
  coalesceLoops.
  
  #### Return modes
  
  The return handle points to the coalesced loop.
  """

  OPERATION_NAME = "transform.loop.coalesce_nested"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def transformed(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(LoopCoalesceNestedOp)
class LoopCoalesceNestedOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.coalesce_nested"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def loop_coalesce_nested(transformed: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return LoopCoalesceNestedOp(transformed=transformed, target=target, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class LoopCoalesceOp(_ods_ir.OpView):
  r"""
  Given a perfect loop nest identified by the outermost loop,
  perform loop coalescing in a bottom-up one-by-one manner.
  
  #### Return modes
  
  The return handle points to the coalesced loop if coalescing happens, or
  the given input loop if coalescing does not happen.
  """

  OPERATION_NAME = "transform.loop.coalesce"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def transformed(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(LoopCoalesceOp)
class LoopCoalesceOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.coalesce"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def loop_coalesce(transformed: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return LoopCoalesceOp(transformed=transformed, target=target, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class LoopFuseSiblingOp(_ods_ir.OpView):
  r"""
  Fuses the `target` loop into the `source` loop assuming they are
  independent of each other. In the fused loop, the arguments, body and
  results of `target` are placed _before_ those of `source`.
  
  For fusion of two `scf.for` loops, the bounds and step size must match. For
  fusion of two `scf.forall` loops, the bounds and the mapping must match.
  Otherwise a silencable failure is produced.
  
  The `target` and `source` handles must refer to exactly one operation,
  otherwise a definite failure is produced. It is the responsibility of the
  user to ensure that the `target` and `source` loops are independent of each
  other -- this op will only perform rudimentary legality checks.
  
  #### Return modes
  
  This operation consumes the `target` and `source` handles and produces the
  `fused_loop` handle, which points to the fused loop.
  """

  OPERATION_NAME = "transform.loop.fuse_sibling"

  _ODS_REGIONS = (0, True)

  def __init__(self, fused_loop: _ods_ir.Type, target: _ods_ir.Value, source: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    operands.append(source)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(fused_loop)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def source(self) -> _ods_ir.Value:
    return self.operation.operands[1]

  @builtins.property
  def fused_loop(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(LoopFuseSiblingOp)
class LoopFuseSiblingOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.fuse_sibling"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def source(self) -> _ods_ir.Value:
    return self.operands[1]

def loop_fuse_sibling(fused_loop: _ods_ir.Type, target: _ods_ir.Value, source: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return LoopFuseSiblingOp(fused_loop=fused_loop, target=target, source=source, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class LoopOutlineOp(_ods_ir.OpView):
  r"""
  Moves the loop into a separate function with the specified name and replaces
  the loop in the Payload IR with a call to that function. Takes care of
  forwarding values that are used in the loop as function arguments. If the
  operand is associated with more than one loop, each loop will be outlined
  into a separate function. The provided name is used as a _base_ for forming
  actual function names following `SymbolTable` auto-renaming scheme to avoid
  duplicate symbols. Expects that all ops in the Payload IR have a
  `SymbolTable` ancestor (typically true because of the top-level module).
  
  #### Return Modes
  
  Returns a handle to the list of outlined functions and a handle to the
  corresponding function call operations in the same order as the operand
  handle.
  
  Produces a definite failure if outlining failed for any of the targets.
  """

  OPERATION_NAME = "transform.loop.outline"

  _ODS_REGIONS = (0, True)

  def __init__(self, function: _ods_ir.Type, call: _ods_ir.Type, target: _ods_ir.Value, func_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["func_name"] = (func_name if (
    isinstance(func_name, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrAttr')) else
      _ods_ir.AttrBuilder.get('StrAttr')(func_name, context=_ods_context))
    results = []
    results.append(function)
    results.append(call)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def func_name(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["func_name"]

  @func_name.setter
  def func_name(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["func_name"] = value

  @builtins.property
  def function(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

  @builtins.property
  def call(self) -> _ods_ir.OpResult:
    return self.operation.results[1]

@_ods_cext.register_op_adaptor(LoopOutlineOp)
class LoopOutlineOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.outline"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def func_name(self) -> _ods_ir.StringAttr:
    return self.attributes["func_name"]

def loop_outline(function: _ods_ir.Type, call: _ods_ir.Type, target: _ods_ir.Value, func_name: _Union[str, _ods_ir.StringAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResultList:
  return LoopOutlineOp(function=function, call=call, target=target, func_name=func_name, loc=loc, ip=ip).results

@_ods_cext.register_operation(_Dialect)
class LoopPeelOp(_ods_ir.OpView):
  r"""
  Rewrite the given loop with a main loop and a partial (first or last) loop.
  When the `peelFront` option is set to true, the first iteration is peeled off.
  Otherwise, updates the given loop so that its step evenly divides its range and puts
  the remaining iteration into a separate loop or a conditional.
  
  In the absence of sufficient static information, this op may peel a loop,
  even if the step always divides the range evenly at runtime.
  
  #### Return modes
  
  This operation ignores non-scf::ForOp ops and drops them in the return.
  The op returns two loops, the peeled loop which has trip count divisible
  by the step, and the remainder loop.
  
  When `peelFront` is true, the first result (remainder loop) executes all
  but the first iteration of the target loop. The second result (peeled
  loop) corresponds to the first iteration of the loop which can be
  canonicalized away in the following optimizations.
  
  When `peelFront` is false, the first result (peeled loop) is the portion
  of the target loop with the highest upper bound that is divisible by the
  step. The second result (remainder loop) contains the remaining iterations. 
  
  Note that even though the Payload IR modification may be performed
  in-place, this operation consumes the operand handle and produces a new one.
  
  #### Return Modes
  
  Produces a definite failure if peeling fails.
  """

  OPERATION_NAME = "transform.loop.peel"

  _ODS_REGIONS = (0, True)

  def __init__(self, peeled_loop: _ods_ir.Type, remainder_loop: _ods_ir.Type, target: _ods_ir.Value, *, peel_front: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, fail_if_already_divisible: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    if peel_front is not None: attributes["peel_front"] = (peel_front if (
        isinstance(peel_front, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(peel_front, context=_ods_context))
    if fail_if_already_divisible is not None: attributes["fail_if_already_divisible"] = (fail_if_already_divisible if (
        isinstance(fail_if_already_divisible, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(fail_if_already_divisible, context=_ods_context))
    results = []
    results.append(peeled_loop)
    results.append(remainder_loop)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def peel_front(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["peel_front"]

  @peel_front.setter
  def peel_front(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["peel_front"] = value

  @builtins.property
  def fail_if_already_divisible(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["fail_if_already_divisible"]

  @fail_if_already_divisible.setter
  def fail_if_already_divisible(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["fail_if_already_divisible"] = value

  @builtins.property
  def peeled_loop(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

  @builtins.property
  def remainder_loop(self) -> _ods_ir.OpResult:
    return self.operation.results[1]

@_ods_cext.register_op_adaptor(LoopPeelOp)
class LoopPeelOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.peel"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def peel_front(self) -> _ods_ir.BoolAttr:
    return self.attributes["peel_front"]

  @builtins.property
  def fail_if_already_divisible(self) -> _ods_ir.BoolAttr:
    return self.attributes["fail_if_already_divisible"]

def loop_peel(peeled_loop: _ods_ir.Type, remainder_loop: _ods_ir.Type, target: _ods_ir.Value, *, peel_front: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, fail_if_already_divisible: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResultList:
  return LoopPeelOp(peeled_loop=peeled_loop, remainder_loop=remainder_loop, target=target, peel_front=peel_front, fail_if_already_divisible=fail_if_already_divisible, loc=loc, ip=ip).results

@_ods_cext.register_operation(_Dialect)
class LoopPipelineOp(_ods_ir.OpView):
  r"""
  Transforms the given loops one by one to achieve software pipelining for
  each of them. That is, performs some amount of reads from memory before the
  loop rather than inside the loop, the same amount of writes into memory
  after the loop, and updates each iteration to read the data for a following
  iteration rather than the current one.
  
  The amount is specified by the attributes.
  
  The values read and about to be stored are transferred as loop iteration
  arguments. Currently supports memref and vector transfer operations as
  memory reads/writes.
  
  #### Return modes
  
  This operation ignores non-scf::For ops and drops them in the return.
  If all the operations referred to by the `target` PDLOperation pipeline
  properly, the transform succeeds. Otherwise the transform produces a
  silenceable failure.  The return handle points to only the subset of
  successfully produced pipelined loops, which can be empty.
  """

  OPERATION_NAME = "transform.loop.pipeline"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _ods_ir.Type, target: _ods_ir.Value, *, iteration_interval: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, read_latency: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    if iteration_interval is not None: attributes["iteration_interval"] = (iteration_interval if (
        isinstance(iteration_interval, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I64Attr')) else
          _ods_ir.AttrBuilder.get('I64Attr')(iteration_interval, context=_ods_context))
    if read_latency is not None: attributes["read_latency"] = (read_latency if (
        isinstance(read_latency, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('I64Attr')) else
          _ods_ir.AttrBuilder.get('I64Attr')(read_latency, context=_ods_context))
    results = []
    results.append(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def iteration_interval(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["iteration_interval"]

  @iteration_interval.setter
  def iteration_interval(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["iteration_interval"] = value

  @builtins.property
  def read_latency(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["read_latency"]

  @read_latency.setter
  def read_latency(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["read_latency"] = value

  @builtins.property
  def transformed(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(LoopPipelineOp)
class LoopPipelineOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.pipeline"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def iteration_interval(self) -> _ods_ir.IntegerAttr:
    return self.attributes["iteration_interval"]

  @builtins.property
  def read_latency(self) -> _ods_ir.IntegerAttr:
    return self.attributes["read_latency"]

def loop_pipeline(transformed: _ods_ir.Type, target: _ods_ir.Value, *, iteration_interval: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, read_latency: _Optional[_Union[int, _ods_ir.IntegerAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return LoopPipelineOp(transformed=transformed, target=target, iteration_interval=iteration_interval, read_latency=read_latency, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class LoopPromoteIfOneIterationOp(_ods_ir.OpView):
  r"""
  Promotes the given target loop op if it has a single iteration. I.e., the
  loop op is removed and only the body remains.
  
  #### Return modes
  
  This transform fails if the target is mapped to ops that are loops. Ops are
  considered loops if they implement the `LoopLikeOpInterface`. Otherwise,
  this transform always succeeds. The transform consumes the target handle and
  modifies the payload.
  """

  OPERATION_NAME = "transform.loop.promote_if_one_iteration"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(LoopPromoteIfOneIterationOp)
class LoopPromoteIfOneIterationOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.promote_if_one_iteration"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def loop_promote_if_one_iteration(target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> LoopPromoteIfOneIterationOp:
  return LoopPromoteIfOneIterationOp(target=target, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class LoopUnrollAndJamOp(_ods_ir.OpView):
  r"""
  Unrolls & jams each loop associated with the given handle to have up to the given
  number of loop body copies per iteration. If the unroll factor is larger
  than the loop trip count, the latter is used as the unroll factor instead.
  
  #### Return modes
  
  This operation ignores non-`scf.for`, non-`affine.for` ops and drops them
  in the return. If all the operations referred to by the `target` operand
  unroll properly, the transform succeeds. Otherwise the transform produces a
  silenceable failure.
  
  Does not return handles as the operation may result in the loop being
  removed after a full unrolling.
  """

  OPERATION_NAME = "transform.loop.unroll_and_jam"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, factor: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["factor"] = (factor if (
    isinstance(factor, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(factor, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def factor(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["factor"]

  @factor.setter
  def factor(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["factor"] = value

@_ods_cext.register_op_adaptor(LoopUnrollAndJamOp)
class LoopUnrollAndJamOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.unroll_and_jam"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def factor(self) -> _ods_ir.IntegerAttr:
    return self.attributes["factor"]

def loop_unroll_and_jam(target: _ods_ir.Value, factor: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> LoopUnrollAndJamOp:
  return LoopUnrollAndJamOp(target=target, factor=factor, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class LoopUnrollOp(_ods_ir.OpView):
  r"""
  Unrolls each loop associated with the given handle to have up to the given
  number of loop body copies per iteration. If the unroll factor is larger
  than the loop trip count, the latter is used as the unroll factor instead.
  
  #### Return modes
  
  This operation ignores non-`scf.for`, non-`affine.for` ops and drops them
  in the return. If all the operations referred to by the `target` operand
  unroll properly, the transform succeeds. Otherwise the transform produces a
  silenceable failure.
  
  Does not return handles as the operation may result in the loop being
  removed after a full unrolling.
  """

  OPERATION_NAME = "transform.loop.unroll"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, factor: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["factor"] = (factor if (
    isinstance(factor, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(factor, context=_ods_context))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def factor(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["factor"]

  @factor.setter
  def factor(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["factor"] = value

@_ods_cext.register_op_adaptor(LoopUnrollOp)
class LoopUnrollOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.unroll"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def factor(self) -> _ods_ir.IntegerAttr:
    return self.attributes["factor"]

def loop_unroll(target: _ods_ir.Value, factor: _Union[int, _ods_ir.IntegerAttr], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> LoopUnrollOp:
  return LoopUnrollOp(target=target, factor=factor, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class ParallelForToNestedForOps(_ods_ir.OpView):
  r"""
  Converts the `scf.parallel` operation pointed to by the given handle into a
  set of nested `scf.for` operations. Each new operation corresponds to one
  dimension of the original parallel loop.
  
  The operand handle must be associated with exactly one payload operation.
  
  Loops with shared outputs are currently not supported.
  
  #### Return Modes
  
  Consumes the operand handle. Produces a silenceable failure if the operand
  is not associated with a single `scf.parallel` payload operation.
  Returns as many handles as the given `parallel` op has dimensions that are
  associated with the generated `scf.for` loops.
  Produces a silenceable failure if another number of resulting handles is
  requested.
  """

  OPERATION_NAME = "transform.loop.parallel_for_to_nested_fors"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _Sequence[_ods_ir.Type], target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.extend(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def transformed(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

@_ods_cext.register_op_adaptor(ParallelForToNestedForOps)
class ParallelForToNestedForOpsAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.loop.parallel_for_to_nested_fors"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def loop_parallel_for_to_nested_fors(transformed: _Sequence[_ods_ir.Type], target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ParallelForToNestedForOps]:
  op = ParallelForToNestedForOps(transformed=transformed, target=target, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)

@_ods_cext.register_operation(_Dialect)
class TakeAssumedBranchOp(_ods_ir.OpView):
  r"""
  Given an scf.if conditional, inject user-defined information that it is
  always safe to execute only the if or else branch.
  
  This is achieved by just replacing the scf.if by the content of one of its
  branches.
  
  This is particularly useful for user-controlled rewriting of conditionals
  that exist solely to guard against out-of-bounds behavior.
  
  At the moment, no assume or assert operation is emitted as it is not always
  desirable. In the future, this may be controlled by a dedicated attribute.
  
  #### Return modes
  
  The transform only consumes its operand and does not produce any result.
  The transform definitely fails if `take_else_branch` is specified and the
  `else` region is empty.
  """

  OPERATION_NAME = "transform.scf.take_assumed_branch"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, *, take_else_branch: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    if bool(take_else_branch): attributes["take_else_branch"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def take_else_branch(self) -> bool:
    return "take_else_branch" in self.operation.attributes

  @take_else_branch.setter
  def take_else_branch(self, value):
    if bool(value):
      self.operation.attributes["take_else_branch"] = _ods_ir.UnitAttr.get()
    elif "take_else_branch" in self.operation.attributes:
      del self.operation.attributes["take_else_branch"]

  @take_else_branch.deleter
  def take_else_branch(self):
    del self.operation.attributes["take_else_branch"]

@_ods_cext.register_op_adaptor(TakeAssumedBranchOp)
class TakeAssumedBranchOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.scf.take_assumed_branch"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def take_else_branch(self) -> bool:
    return "take_else_branch" in self.attributes

def scf_take_assumed_branch(target: _ods_ir.Value, *, take_else_branch: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> TakeAssumedBranchOp:
  return TakeAssumedBranchOp(target=target, take_else_branch=take_else_branch, loc=loc, ip=ip)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Typing support for CUDA vector types (float32x4, int32x2, etc.)
"""

import operator
from numba_cuda_mlir.lowering_utilities.type_conversions import float_of_width
from numba_cuda_mlir.numba_cuda.typing.templates import (
    AbstractTemplate,
    AttributeTemplate,
    Registry,
    signature,
)
from numba_cuda_mlir import types
from numba_cuda_mlir.type_defs.vector_types import VectorType
from numba_cuda_mlir.cuda.vector_types import (
    _vector_types,
)

from numba_cuda_mlir.numba_cuda.types import Callable, DTypeSpec
from numba_cuda_mlir.numba_cuda.typing import typeof

registry = Registry()


class VectorTypeClass(Callable, DTypeSpec):
    """
    A class representing a vector type that can be called to construct instances
    and used as a dtype specifier.
    """

    def __init__(self, instance_type, constructor_template):
        self.instance_type = instance_type
        self._template = constructor_template
        self.typing_key = instance_type
        super().__init__(f"class({instance_type})")

    @property
    def dtype(self):
        return self.instance_type

    def get_call_type(self, context, args, kws):
        return self._template(context).apply(args, kws)

    def get_call_signatures(self):
        sigs = getattr(self._template, "cases", [])
        is_param = hasattr(self._template, "generic")
        return sigs, is_param

    def get_impl_key(self, sig):
        return self.typing_key


# Attribute index mapping
ATTR_INDEX = {"x": 0, "y": 1, "z": 2, "w": 3}


_constructor_template_cache = {}


def make_constructor_template(vec_type):
    """Create a typing template for a vector type constructor (cached)."""
    if vec_type in _constructor_template_cache:
        return _constructor_template_cache[vec_type]

    num_elements = vec_type.length

    class ConstructorTemplate(AbstractTemplate):
        key = vec_type

        def generic(self, args, kws):
            if kws:
                return None

            # Single argument cases (must check before multi-arg scalar case)
            if len(args) == 1:
                arg = args[0]
                # Copy from compatible vector (same number of elements)
                if isinstance(arg, VectorType):
                    if arg.length == num_elements:
                        return signature(vec_type, arg)
                # Scalar broadcast
                if isinstance(arg, (types.Integer, types.Float)):
                    return signature(vec_type, arg)
                # Complex broadcast/cast
                if isinstance(arg, types.Complex) and num_elements == 2:
                    return signature(vec_type, arg)

            # All scalar arguments matching element count
            if len(args) == num_elements:
                all_scalars = all(
                    isinstance(arg, (types.Integer, types.Float, types.Boolean)) for arg in args
                )
                if all_scalars:
                    return signature(vec_type, *args)

            # Mixed vector/scalar arguments
            total_elements = 0
            for arg in args:
                if isinstance(arg, VectorType):
                    total_elements += arg.length
                elif isinstance(arg, (types.Integer, types.Float, types.Boolean)):
                    total_elements += 1
                else:
                    return None

            if total_elements == num_elements:
                return signature(vec_type, *args)

            return None

    ConstructorTemplate.__name__ = f"{vec_type.name}ConstructorTemplate"
    _constructor_template_cache[vec_type] = ConstructorTemplate
    return ConstructorTemplate


# Register all vector type constructors
for vec_type in _vector_types:
    template = make_constructor_template(vec_type)
    registry.register(template)
    registry.register_global(vec_type, VectorTypeClass(vec_type, template))


def _get_vector_type(dtype, length):
    for vt in _vector_types:
        if vt.dtype == dtype and vt.length == length:
            return vt
    return None


def make_vector_binop_template(op):
    class VectorBinOpTemplate(AbstractTemplate):
        key = op

        def generic(self, args, kws):
            if len(args) != 2:
                return None

            lhs, rhs = args

            if isinstance(lhs, VectorType) and isinstance(rhs, VectorType):
                if op not in (operator.add, operator.iadd, operator.sub, operator.isub):
                    return None

                if lhs.length != rhs.length:
                    return None
                target_dtype = self.context.unify_types(lhs.dtype, rhs.dtype)
                if target_dtype is None:
                    return None

                if op == operator.truediv and isinstance(target_dtype, types.Integer):
                    bitwidth = max(target_dtype.bitwidth, 32)
                    target_dtype = float_of_width(bitwidth)

                restype = _get_vector_type(target_dtype, lhs.length)
                if restype is None:
                    return None
                return signature(restype, lhs, rhs)

            elif isinstance(lhs, VectorType) and isinstance(rhs, types.Number):
                target_dtype = self.context.unify_types(lhs.dtype, rhs)
                if target_dtype is None:
                    return None

                if op == operator.truediv and isinstance(target_dtype, types.Integer):
                    bitwidth = max(target_dtype.bitwidth, 32)
                    target_dtype = float_of_width(bitwidth)

                restype = _get_vector_type(target_dtype, lhs.length)
                if restype is None:
                    return None
                return signature(restype, lhs, rhs)

            elif isinstance(lhs, types.Number) and isinstance(rhs, VectorType):
                target_dtype = self.context.unify_types(lhs, rhs.dtype)
                if target_dtype is None:
                    return None

                if op == operator.truediv and isinstance(target_dtype, types.Integer):
                    bitwidth = max(target_dtype.bitwidth, 32)
                    target_dtype = float_of_width(bitwidth)

                restype = _get_vector_type(target_dtype, rhs.length)
                if restype is None:
                    return None
                return signature(restype, lhs, rhs)

            return None

    VectorBinOpTemplate.__name__ = f"VectorBinOpTemplate_{op.__name__}"
    return VectorBinOpTemplate


for op in [
    operator.add,
    operator.iadd,
    operator.sub,
    operator.isub,
    operator.mul,
    operator.imul,
    operator.truediv,
    operator.itruediv,
    operator.floordiv,
    operator.ifloordiv,
    operator.mod,
    operator.imod,
]:
    registry.register_global(op, types.Function(make_vector_binop_template(op)))


@registry.register_global(operator.neg)
class VectorNegTemplate(AbstractTemplate):
    def generic(self, args, kws):
        if len(args) != 1:
            return None
        x = args[0]
        if isinstance(x, VectorType):
            return signature(x, x)
        return None


@registry.register_global(abs)
class VectorAbsTemplate(AbstractTemplate):
    def generic(self, args, kws):
        if len(args) != 1:
            return None
        x = args[0]
        if isinstance(x, VectorType):
            return signature(x, x)
        return None


@registry.register_attr
class VectorTypeAttributeTemplate(AttributeTemplate):
    """Typing for vector type attribute access (.x, .y, .z, .w)."""

    key = VectorType

    def resolve_x(self, ty):
        if ty.length >= 1:
            return ty.dtype

    def resolve_y(self, ty):
        if ty.length >= 2:
            return ty.dtype

    def resolve_z(self, ty):
        if ty.length >= 3:
            return ty.dtype

    def resolve_w(self, ty):
        if ty.length >= 4:
            return ty.dtype


# Register typeof_impl for all vector types so they can be used as closure variables


@typeof.typeof_impl.register(VectorType)
def typeof_vector_type(val, c):
    """Create a typeof implementation for a vector type."""
    template = make_constructor_template(val)
    return VectorTypeClass(val, template)


@registry.register_global(complex)
class ComplexBuiltinTemplate(AbstractTemplate):
    def generic(self, args, kws):
        if len(args) == 1 and isinstance(args[0], VectorType) and args[0].length == 2:
            dtype = args[0].dtype
            if (isinstance(dtype, types.Float) and dtype.bitwidth <= 32) or (
                isinstance(dtype, types.Integer) and dtype.bitwidth <= 16
            ):
                return signature(types.complex64, args[0])
            else:
                return signature(types.complex128, args[0])
        return None



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba_cuda_mlir.types import (
    int16,
    int32,
    int64,
    float32,
    float64,
    UniTuple,
    Tuple,
)


def abs(x: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_abs.html

    CAPI: int32 __nv_abs(int32 x);
    """


def acos(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acos.html

    CAPI: float64 __nv_acos(float64 x);
    """


def acosf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acosf.html

    CAPI: float32 __nv_acosf(float32 x);
    """


def acosh(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acosh.html

    CAPI: float64 __nv_acosh(float64 x);
    """


def acoshf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_acoshf.html

    CAPI: float32 __nv_acoshf(float32 x);
    """


def asin(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asin.html

    CAPI: float64 __nv_asin(float64 x);
    """


def asinf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asinf.html

    CAPI: float32 __nv_asinf(float32 x);
    """


def asinh(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asinh.html

    CAPI: float64 __nv_asinh(float64 x);
    """


def asinhf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_asinhf.html

    CAPI: float32 __nv_asinhf(float32 x);
    """


def atan(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atan.html

    CAPI: float64 __nv_atan(float64 x);
    """


def atan2(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atan2.html

    CAPI: float64 __nv_atan2(float64 x, float64 y);
    """


def atan2f(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atan2f.html

    CAPI: float32 __nv_atan2f(float32 x, float32 y);
    """


def atanf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atanf.html

    CAPI: float32 __nv_atanf(float32 x);
    """


def atanh(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atanh.html

    CAPI: float64 __nv_atanh(float64 x);
    """


def atanhf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_atanhf.html

    CAPI: float32 __nv_atanhf(float32 x);
    """


def brev(x: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_brev.html

    CAPI: int32 __nv_brev(int32 x);
    """


def brevll(x: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_brevll.html

    CAPI: int64 __nv_brevll(int64 x);
    """


def byte_perm(x: int32, y: int32, z: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_byte_perm.html

    CAPI: int32 __nv_byte_perm(int32 x, int32 y, int32 z);
    """


def cbrt(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cbrt.html

    CAPI: float64 __nv_cbrt(float64 x);
    """


def cbrtf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cbrtf.html

    CAPI: float32 __nv_cbrtf(float32 x);
    """


def ceil(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ceil.html

    CAPI: float64 __nv_ceil(float64 x);
    """


def ceilf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ceilf.html

    CAPI: float32 __nv_ceilf(float32 x);
    """


def clz(x: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_clz.html

    CAPI: int32 __nv_clz(int32 x);
    """


def clzll(x: int64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_clzll.html

    CAPI: int32 __nv_clzll(int64 x);
    """


def copysign(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_copysign.html

    CAPI: float64 __nv_copysign(float64 x, float64 y);
    """


def copysignf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_copysignf.html

    CAPI: float32 __nv_copysignf(float32 x, float32 y);
    """


def cos(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cos.html

    CAPI: float64 __nv_cos(float64 x);
    """


def cosf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cosf.html

    CAPI: float32 __nv_cosf(float32 x);
    """


def cosh(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cosh.html

    CAPI: float64 __nv_cosh(float64 x);
    """


def coshf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_coshf.html

    CAPI: float32 __nv_coshf(float32 x);
    """


def cospi(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cospi.html

    CAPI: float64 __nv_cospi(float64 x);
    """


def cospif(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_cospif.html

    CAPI: float32 __nv_cospif(float32 x);
    """


def dadd_rd(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_rd.html

    CAPI: float64 __nv_dadd_rd(float64 x, float64 y);
    """


def dadd_rn(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_rn.html

    CAPI: float64 __nv_dadd_rn(float64 x, float64 y);
    """


def dadd_ru(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_ru.html

    CAPI: float64 __nv_dadd_ru(float64 x, float64 y);
    """


def dadd_rz(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dadd_rz.html

    CAPI: float64 __nv_dadd_rz(float64 x, float64 y);
    """


def ddiv_rd(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_rd.html

    CAPI: float64 __nv_ddiv_rd(float64 x, float64 y);
    """


def ddiv_rn(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_rn.html

    CAPI: float64 __nv_ddiv_rn(float64 x, float64 y);
    """


def ddiv_ru(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_ru.html

    CAPI: float64 __nv_ddiv_ru(float64 x, float64 y);
    """


def ddiv_rz(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ddiv_rz.html

    CAPI: float64 __nv_ddiv_rz(float64 x, float64 y);
    """


def dmul_rd(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_rd.html

    CAPI: float64 __nv_dmul_rd(float64 x, float64 y);
    """


def dmul_rn(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_rn.html

    CAPI: float64 __nv_dmul_rn(float64 x, float64 y);
    """


def dmul_ru(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_ru.html

    CAPI: float64 __nv_dmul_ru(float64 x, float64 y);
    """


def dmul_rz(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dmul_rz.html

    CAPI: float64 __nv_dmul_rz(float64 x, float64 y);
    """


def double2float_rd(d: float64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_rd.html

    CAPI: float32 __nv_double2float_rd(float64 d);
    """


def double2float_rn(d: float64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_rn.html

    CAPI: float32 __nv_double2float_rn(float64 d);
    """


def double2float_ru(d: float64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_ru.html

    CAPI: float32 __nv_double2float_ru(float64 d);
    """


def double2float_rz(d: float64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2float_rz.html

    CAPI: float32 __nv_double2float_rz(float64 d);
    """


def double2hiint(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2hiint.html

    CAPI: int32 __nv_double2hiint(float64 d);
    """


def double2int_rd(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_rd.html

    CAPI: int32 __nv_double2int_rd(float64 d);
    """


def double2int_rn(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_rn.html

    CAPI: int32 __nv_double2int_rn(float64 d);
    """


def double2int_ru(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_ru.html

    CAPI: int32 __nv_double2int_ru(float64 d);
    """


def double2int_rz(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2int_rz.html

    CAPI: int32 __nv_double2int_rz(float64 d);
    """


def double2ll_rd(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_rd.html

    CAPI: int64 __nv_double2ll_rd(float64 f);
    """


def double2ll_rn(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_rn.html

    CAPI: int64 __nv_double2ll_rn(float64 f);
    """


def double2ll_ru(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_ru.html

    CAPI: int64 __nv_double2ll_ru(float64 f);
    """


def double2ll_rz(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ll_rz.html

    CAPI: int64 __nv_double2ll_rz(float64 f);
    """


def double2loint(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2loint.html

    CAPI: int32 __nv_double2loint(float64 d);
    """


def double2uint_rd(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_rd.html

    CAPI: int32 __nv_double2uint_rd(float64 d);
    """


def double2uint_rn(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_rn.html

    CAPI: int32 __nv_double2uint_rn(float64 d);
    """


def double2uint_ru(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_ru.html

    CAPI: int32 __nv_double2uint_ru(float64 d);
    """


def double2uint_rz(d: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2uint_rz.html

    CAPI: int32 __nv_double2uint_rz(float64 d);
    """


def double2ull_rd(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_rd.html

    CAPI: int64 __nv_double2ull_rd(float64 f);
    """


def double2ull_rn(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_rn.html

    CAPI: int64 __nv_double2ull_rn(float64 f);
    """


def double2ull_ru(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_ru.html

    CAPI: int64 __nv_double2ull_ru(float64 f);
    """


def double2ull_rz(f: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double2ull_rz.html

    CAPI: int64 __nv_double2ull_rz(float64 f);
    """


def double_as_longlong(x: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_double_as_longlong.html

    CAPI: int64 __nv_double_as_longlong(float64 x);
    """


def drcp_rd(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_rd.html

    CAPI: float64 __nv_drcp_rd(float64 x);
    """


def drcp_rn(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_rn.html

    CAPI: float64 __nv_drcp_rn(float64 x);
    """


def drcp_ru(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_ru.html

    CAPI: float64 __nv_drcp_ru(float64 x);
    """


def drcp_rz(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_drcp_rz.html

    CAPI: float64 __nv_drcp_rz(float64 x);
    """


def dsqrt_rd(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_rd.html

    CAPI: float64 __nv_dsqrt_rd(float64 x);
    """


def dsqrt_rn(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_rn.html

    CAPI: float64 __nv_dsqrt_rn(float64 x);
    """


def dsqrt_ru(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_ru.html

    CAPI: float64 __nv_dsqrt_ru(float64 x);
    """


def dsqrt_rz(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_dsqrt_rz.html

    CAPI: float64 __nv_dsqrt_rz(float64 x);
    """


def erf(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erf.html

    CAPI: float64 __nv_erf(float64 x);
    """


def erfc(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfc.html

    CAPI: float64 __nv_erfc(float64 x);
    """


def erfcf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcf.html

    CAPI: float32 __nv_erfcf(float32 x);
    """


def erfcinv(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcinv.html

    CAPI: float64 __nv_erfcinv(float64 x);
    """


def erfcinvf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcinvf.html

    CAPI: float32 __nv_erfcinvf(float32 x);
    """


def erfcx(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcx.html

    CAPI: float64 __nv_erfcx(float64 x);
    """


def erfcxf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfcxf.html

    CAPI: float32 __nv_erfcxf(float32 x);
    """


def erff(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erff.html

    CAPI: float32 __nv_erff(float32 x);
    """


def erfinv(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfinv.html

    CAPI: float64 __nv_erfinv(float64 x);
    """


def erfinvf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_erfinvf.html

    CAPI: float32 __nv_erfinvf(float32 x);
    """


def exp(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp.html

    CAPI: float64 __nv_exp(float64 x);
    """


def exp10(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp10.html

    CAPI: float64 __nv_exp10(float64 x);
    """


def exp10f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp10f.html

    CAPI: float32 __nv_exp10f(float32 x);
    """


def exp2(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp2.html

    CAPI: float64 __nv_exp2(float64 x);
    """


def exp2f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_exp2f.html

    CAPI: float32 __nv_exp2f(float32 x);
    """


def expf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_expf.html

    CAPI: float32 __nv_expf(float32 x);
    """


def expm1(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_expm1.html

    CAPI: float64 __nv_expm1(float64 x);
    """


def expm1f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_expm1f.html

    CAPI: float32 __nv_expm1f(float32 x);
    """


def fabs(f: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fabs.html

    CAPI: float64 __nv_fabs(float64 f);
    """


def fabsf(f: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fabsf.html

    CAPI: float32 __nv_fabsf(float32 f);
    """


def fadd_rd(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_rd.html

    CAPI: float32 __nv_fadd_rd(float32 x, float32 y);
    """


def fadd_rn(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_rn.html

    CAPI: float32 __nv_fadd_rn(float32 x, float32 y);
    """


def fadd_ru(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_ru.html

    CAPI: float32 __nv_fadd_ru(float32 x, float32 y);
    """


def fadd_rz(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fadd_rz.html

    CAPI: float32 __nv_fadd_rz(float32 x, float32 y);
    """


def fast_cosf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_cosf.html

    CAPI: float32 __nv_fast_cosf(float32 x);
    """


def fast_exp10f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_exp10f.html

    CAPI: float32 __nv_fast_exp10f(float32 x);
    """


def fast_expf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_expf.html

    CAPI: float32 __nv_fast_expf(float32 x);
    """


def fast_fdividef(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_fdividef.html

    CAPI: float32 __nv_fast_fdividef(float32 x, float32 y);
    """


def fast_log10f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_log10f.html

    CAPI: float32 __nv_fast_log10f(float32 x);
    """


def fast_log2f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_log2f.html

    CAPI: float32 __nv_fast_log2f(float32 x);
    """


def fast_logf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_logf.html

    CAPI: float32 __nv_fast_logf(float32 x);
    """


def fast_powf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_powf.html

    CAPI: float32 __nv_fast_powf(float32 x, float32 y);
    """


def fast_sincosf(x: float32) -> UniTuple(float32, 2):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_sincosf.html

    CAPI: void __nv_fast_sincosf(float32 x, float32* sptr, float32* cptr);
    """


def fast_sinf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_sinf.html

    CAPI: float32 __nv_fast_sinf(float32 x);
    """


def fast_tanf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fast_tanf.html

    CAPI: float32 __nv_fast_tanf(float32 x);
    """


def fdim(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdim.html

    CAPI: float64 __nv_fdim(float64 x, float64 y);
    """


def fdimf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdimf.html

    CAPI: float32 __nv_fdimf(float32 x, float32 y);
    """


def fdiv_rd(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_rd.html

    CAPI: float32 __nv_fdiv_rd(float32 x, float32 y);
    """


def fdiv_rn(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_rn.html

    CAPI: float32 __nv_fdiv_rn(float32 x, float32 y);
    """


def fdiv_ru(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_ru.html

    CAPI: float32 __nv_fdiv_ru(float32 x, float32 y);
    """


def fdiv_rz(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fdiv_rz.html

    CAPI: float32 __nv_fdiv_rz(float32 x, float32 y);
    """


def ffs(x: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ffs.html

    CAPI: int32 __nv_ffs(int32 x);
    """


def ffsll(x: int64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ffsll.html

    CAPI: int32 __nv_ffsll(int64 x);
    """


def finitef(x: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_finitef.html

    CAPI: int32 __nv_finitef(float32 x);
    """


def float2half_rn(f: float32) -> int16:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2half_rn.html

    CAPI: int16 __nv_float2half_rn(float32 f);
    """


def float2int_rd(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_rd.html

    CAPI: int32 __nv_float2int_rd(float32 in);
    """


def float2int_rn(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_rn.html

    CAPI: int32 __nv_float2int_rn(float32 in);
    """


def float2int_ru(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_ru.html

    CAPI: int32 __nv_float2int_ru(float32 in);
    """


def float2int_rz(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2int_rz.html

    CAPI: int32 __nv_float2int_rz(float32 in);
    """


def float2ll_rd(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_rd.html

    CAPI: int64 __nv_float2ll_rd(float32 f);
    """


def float2ll_rn(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_rn.html

    CAPI: int64 __nv_float2ll_rn(float32 f);
    """


def float2ll_ru(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_ru.html

    CAPI: int64 __nv_float2ll_ru(float32 f);
    """


def float2ll_rz(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ll_rz.html

    CAPI: int64 __nv_float2ll_rz(float32 f);
    """


def float2uint_rd(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_rd.html

    CAPI: int32 __nv_float2uint_rd(float32 in);
    """


def float2uint_rn(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_rn.html

    CAPI: int32 __nv_float2uint_rn(float32 in);
    """


def float2uint_ru(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_ru.html

    CAPI: int32 __nv_float2uint_ru(float32 in);
    """


def float2uint_rz(in_: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2uint_rz.html

    CAPI: int32 __nv_float2uint_rz(float32 in);
    """


def float2ull_rd(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_rd.html

    CAPI: int64 __nv_float2ull_rd(float32 f);
    """


def float2ull_rn(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_rn.html

    CAPI: int64 __nv_float2ull_rn(float32 f);
    """


def float2ull_ru(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_ru.html

    CAPI: int64 __nv_float2ull_ru(float32 f);
    """


def float2ull_rz(f: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float2ull_rz.html

    CAPI: int64 __nv_float2ull_rz(float32 f);
    """


def float_as_int(x: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_float_as_int.html

    CAPI: int32 __nv_float_as_int(float32 x);
    """


def floor(f: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_floor.html

    CAPI: float64 __nv_floor(float64 f);
    """


def floorf(f: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_floorf.html

    CAPI: float32 __nv_floorf(float32 f);
    """


def fma(x: float64, y: float64, z: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma.html

    CAPI: float64 __nv_fma(float64 x, float64 y, float64 z);
    """


def fma_rd(x: float64, y: float64, z: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_rd.html

    CAPI: float64 __nv_fma_rd(float64 x, float64 y, float64 z);
    """


def fma_rn(x: float64, y: float64, z: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_rn.html

    CAPI: float64 __nv_fma_rn(float64 x, float64 y, float64 z);
    """


def fma_ru(x: float64, y: float64, z: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_ru.html

    CAPI: float64 __nv_fma_ru(float64 x, float64 y, float64 z);
    """


def fma_rz(x: float64, y: float64, z: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fma_rz.html

    CAPI: float64 __nv_fma_rz(float64 x, float64 y, float64 z);
    """


def fmaf(x: float32, y: float32, z: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf.html

    CAPI: float32 __nv_fmaf(float32 x, float32 y, float32 z);
    """


def fmaf_rd(x: float32, y: float32, z: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_rd.html

    CAPI: float32 __nv_fmaf_rd(float32 x, float32 y, float32 z);
    """


def fmaf_rn(x: float32, y: float32, z: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_rn.html

    CAPI: float32 __nv_fmaf_rn(float32 x, float32 y, float32 z);
    """


def fmaf_ru(x: float32, y: float32, z: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_ru.html

    CAPI: float32 __nv_fmaf_ru(float32 x, float32 y, float32 z);
    """


def fmaf_rz(x: float32, y: float32, z: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaf_rz.html

    CAPI: float32 __nv_fmaf_rz(float32 x, float32 y, float32 z);
    """


def fmax(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmax.html

    CAPI: float64 __nv_fmax(float64 x, float64 y);
    """


def fmaxf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmaxf.html

    CAPI: float32 __nv_fmaxf(float32 x, float32 y);
    """


def fmin(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmin.html

    CAPI: float64 __nv_fmin(float64 x, float64 y);
    """


def fminf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fminf.html

    CAPI: float32 __nv_fminf(float32 x, float32 y);
    """


def fmod(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmod.html

    CAPI: float64 __nv_fmod(float64 x, float64 y);
    """


def fmodf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmodf.html

    CAPI: float32 __nv_fmodf(float32 x, float32 y);
    """


def fmul_rd(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_rd.html

    CAPI: float32 __nv_fmul_rd(float32 x, float32 y);
    """


def fmul_rn(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_rn.html

    CAPI: float32 __nv_fmul_rn(float32 x, float32 y);
    """


def fmul_ru(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_ru.html

    CAPI: float32 __nv_fmul_ru(float32 x, float32 y);
    """


def fmul_rz(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fmul_rz.html

    CAPI: float32 __nv_fmul_rz(float32 x, float32 y);
    """


def frcp_rd(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_rd.html

    CAPI: float32 __nv_frcp_rd(float32 x);
    """


def frcp_rn(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_rn.html

    CAPI: float32 __nv_frcp_rn(float32 x);
    """


def frcp_ru(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_ru.html

    CAPI: float32 __nv_frcp_ru(float32 x);
    """


def frcp_rz(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frcp_rz.html

    CAPI: float32 __nv_frcp_rz(float32 x);
    """


def frexp(x: float64) -> Tuple([float64, int32]):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frexp.html

    CAPI: float64 __nv_frexp(float64 x, int32* b);
    """


def frexpf(x: float32) -> Tuple([float32, int32]):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frexpf.html

    CAPI: float32 __nv_frexpf(float32 x, int32* b);
    """


def frsqrt_rn(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_frsqrt_rn.html

    CAPI: float32 __nv_frsqrt_rn(float32 x);
    """


def fsqrt_rd(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_rd.html

    CAPI: float32 __nv_fsqrt_rd(float32 x);
    """


def fsqrt_rn(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_rn.html

    CAPI: float32 __nv_fsqrt_rn(float32 x);
    """


def fsqrt_ru(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_ru.html

    CAPI: float32 __nv_fsqrt_ru(float32 x);
    """


def fsqrt_rz(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsqrt_rz.html

    CAPI: float32 __nv_fsqrt_rz(float32 x);
    """


def fsub_rd(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_rd.html

    CAPI: float32 __nv_fsub_rd(float32 x, float32 y);
    """


def fsub_rn(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_rn.html

    CAPI: float32 __nv_fsub_rn(float32 x, float32 y);
    """


def fsub_ru(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_ru.html

    CAPI: float32 __nv_fsub_ru(float32 x, float32 y);
    """


def fsub_rz(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_fsub_rz.html

    CAPI: float32 __nv_fsub_rz(float32 x, float32 y);
    """


def hadd(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hadd.html

    CAPI: int32 __nv_hadd(int32 x, int32 y);
    """


def half2float(h: int16) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_half2float.html

    CAPI: float32 __nv_half2float(int16 h);
    """


def hiloint2double(x: int32, y: int32) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hiloint2double.html

    CAPI: float64 __nv_hiloint2double(int32 x, int32 y);
    """


def hypot(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hypot.html

    CAPI: float64 __nv_hypot(float64 x, float64 y);
    """


def hypotf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_hypotf.html

    CAPI: float32 __nv_hypotf(float32 x, float32 y);
    """


def ilogb(x: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ilogb.html

    CAPI: int32 __nv_ilogb(float64 x);
    """


def ilogbf(x: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ilogbf.html

    CAPI: int32 __nv_ilogbf(float32 x);
    """


def int2double_rn(i: int32) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2double_rn.html

    CAPI: float64 __nv_int2double_rn(int32 i);
    """


def int2float_rd(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_rd.html

    CAPI: float32 __nv_int2float_rd(int32 in);
    """


def int2float_rn(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_rn.html

    CAPI: float32 __nv_int2float_rn(int32 in);
    """


def int2float_ru(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_ru.html

    CAPI: float32 __nv_int2float_ru(int32 in);
    """


def int2float_rz(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int2float_rz.html

    CAPI: float32 __nv_int2float_rz(int32 in);
    """


def int_as_float(x: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_int_as_float.html

    CAPI: float32 __nv_int_as_float(int32 x);
    """


def isfinited(x: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isfinited.html

    CAPI: int32 __nv_isfinited(float64 x);
    """


def isinfd(x: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isinfd.html

    CAPI: int32 __nv_isinfd(float64 x);
    """


def isinff(x: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isinff.html

    CAPI: int32 __nv_isinff(float32 x);
    """


def isnand(x: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isnand.html

    CAPI: int32 __nv_isnand(float64 x);
    """


def isnanf(x: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_isnanf.html

    CAPI: int32 __nv_isnanf(float32 x);
    """


def j0(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j0.html

    CAPI: float64 __nv_j0(float64 x);
    """


def j0f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j0f.html

    CAPI: float32 __nv_j0f(float32 x);
    """


def j1(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j1.html

    CAPI: float64 __nv_j1(float64 x);
    """


def j1f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_j1f.html

    CAPI: float32 __nv_j1f(float32 x);
    """


def jn(n: int32, x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_jn.html

    CAPI: float64 __nv_jn(int32 n, float64 x);
    """


def jnf(n: int32, x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_jnf.html

    CAPI: float32 __nv_jnf(int32 n, float32 x);
    """


def ldexp(x: float64, y: int32) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ldexp.html

    CAPI: float64 __nv_ldexp(float64 x, int32 y);
    """


def ldexpf(x: float32, y: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ldexpf.html

    CAPI: float32 __nv_ldexpf(float32 x, int32 y);
    """


def lgamma(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_lgamma.html

    CAPI: float64 __nv_lgamma(float64 x);
    """


def lgammaf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_lgammaf.html

    CAPI: float32 __nv_lgammaf(float32 x);
    """


def ll2double_rd(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_rd.html

    CAPI: float64 __nv_ll2double_rd(int64 l);
    """


def ll2double_rn(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_rn.html

    CAPI: float64 __nv_ll2double_rn(int64 l);
    """


def ll2double_ru(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_ru.html

    CAPI: float64 __nv_ll2double_ru(int64 l);
    """


def ll2double_rz(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2double_rz.html

    CAPI: float64 __nv_ll2double_rz(int64 l);
    """


def ll2float_rd(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_rd.html

    CAPI: float32 __nv_ll2float_rd(int64 l);
    """


def ll2float_rn(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_rn.html

    CAPI: float32 __nv_ll2float_rn(int64 l);
    """


def ll2float_ru(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_ru.html

    CAPI: float32 __nv_ll2float_ru(int64 l);
    """


def ll2float_rz(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ll2float_rz.html

    CAPI: float32 __nv_ll2float_rz(int64 l);
    """


def llabs(x: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llabs.html

    CAPI: int64 __nv_llabs(int64 x);
    """


def llmax(x: int64, y: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llmax.html

    CAPI: int64 __nv_llmax(int64 x, int64 y);
    """


def llmin(x: int64, y: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llmin.html

    CAPI: int64 __nv_llmin(int64 x, int64 y);
    """


def llrint(x: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llrint.html

    CAPI: int64 __nv_llrint(float64 x);
    """


def llrintf(x: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llrintf.html

    CAPI: int64 __nv_llrintf(float32 x);
    """


def llround(x: float64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llround.html

    CAPI: int64 __nv_llround(float64 x);
    """


def llroundf(x: float32) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_llroundf.html

    CAPI: int64 __nv_llroundf(float32 x);
    """


def log(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log.html

    CAPI: float64 __nv_log(float64 x);
    """


def log10(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log10.html

    CAPI: float64 __nv_log10(float64 x);
    """


def log10f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log10f.html

    CAPI: float32 __nv_log10f(float32 x);
    """


def log1p(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log1p.html

    CAPI: float64 __nv_log1p(float64 x);
    """


def log1pf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log1pf.html

    CAPI: float32 __nv_log1pf(float32 x);
    """


def log2(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log2.html

    CAPI: float64 __nv_log2(float64 x);
    """


def log2f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_log2f.html

    CAPI: float32 __nv_log2f(float32 x);
    """


def logb(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_logb.html

    CAPI: float64 __nv_logb(float64 x);
    """


def logbf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_logbf.html

    CAPI: float32 __nv_logbf(float32 x);
    """


def logf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_logf.html

    CAPI: float32 __nv_logf(float32 x);
    """


def longlong_as_double(x: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_longlong_as_double.html

    CAPI: float64 __nv_longlong_as_double(int64 x);
    """


def max(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_max.html

    CAPI: int32 __nv_max(int32 x, int32 y);
    """


def min(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_min.html

    CAPI: int32 __nv_min(int32 x, int32 y);
    """


def modf(x: float64) -> UniTuple(float64, 2):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_modf.html

    CAPI: float64 __nv_modf(float64 x, float64* b);
    """


def modff(x: float32) -> UniTuple(float32, 2):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_modff.html

    CAPI: float32 __nv_modff(float32 x, float32* b);
    """


def mul24(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_mul24.html

    CAPI: int32 __nv_mul24(int32 x, int32 y);
    """


def mul64hi(x: int64, y: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_mul64hi.html

    CAPI: int64 __nv_mul64hi(int64 x, int64 y);
    """


def mulhi(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_mulhi.html

    CAPI: int32 __nv_mulhi(int32 x, int32 y);
    """


def nearbyint(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nearbyint.html

    CAPI: float64 __nv_nearbyint(float64 x);
    """


def nearbyintf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nearbyintf.html

    CAPI: float32 __nv_nearbyintf(float32 x);
    """


def nextafter(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nextafter.html

    CAPI: float64 __nv_nextafter(float64 x, float64 y);
    """


def nextafterf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_nextafterf.html

    CAPI: float32 __nv_nextafterf(float32 x, float32 y);
    """


def normcdf(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdf.html

    CAPI: float64 __nv_normcdf(float64 x);
    """


def normcdff(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdff.html

    CAPI: float32 __nv_normcdff(float32 x);
    """


def normcdfinv(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdfinv.html

    CAPI: float64 __nv_normcdfinv(float64 x);
    """


def normcdfinvf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_normcdfinvf.html

    CAPI: float32 __nv_normcdfinvf(float32 x);
    """


def popc(x: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_popc.html

    CAPI: int32 __nv_popc(int32 x);
    """


def popcll(x: int64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_popcll.html

    CAPI: int32 __nv_popcll(int64 x);
    """


def pow(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_pow.html

    CAPI: float64 __nv_pow(float64 x, float64 y);
    """


def powf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_powf.html

    CAPI: float32 __nv_powf(float32 x, float32 y);
    """


def powi(x: float64, y: int32) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_powi.html

    CAPI: float64 __nv_powi(float64 x, int32 y);
    """


def powif(x: float32, y: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_powif.html

    CAPI: float32 __nv_powif(float32 x, int32 y);
    """


def rcbrt(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rcbrt.html

    CAPI: float64 __nv_rcbrt(float64 x);
    """


def rcbrtf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rcbrtf.html

    CAPI: float32 __nv_rcbrtf(float32 x);
    """


def remainder(x: float64, y: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remainder.html

    CAPI: float64 __nv_remainder(float64 x, float64 y);
    """


def remainderf(x: float32, y: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remainderf.html

    CAPI: float32 __nv_remainderf(float32 x, float32 y);
    """


def remquo(x: float64, y: float64) -> Tuple([float64, int32]):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remquo.html

    CAPI: float64 __nv_remquo(float64 x, float64 y, int32* c);
    """


def remquof(x: float32, y: float32) -> Tuple([float32, int32]):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_remquof.html

    CAPI: float32 __nv_remquof(float32 x, float32 y, int32* quo);
    """


def rhadd(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rhadd.html

    CAPI: int32 __nv_rhadd(int32 x, int32 y);
    """


def rint(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rint.html

    CAPI: float64 __nv_rint(float64 x);
    """


def rintf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rintf.html

    CAPI: float32 __nv_rintf(float32 x);
    """


def round(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_round.html

    CAPI: float64 __nv_round(float64 x);
    """


def roundf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_roundf.html

    CAPI: float32 __nv_roundf(float32 x);
    """


def rsqrt(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rsqrt.html

    CAPI: float64 __nv_rsqrt(float64 x);
    """


def rsqrtf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_rsqrtf.html

    CAPI: float32 __nv_rsqrtf(float32 x);
    """


def sad(x: int32, y: int32, z: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sad.html

    CAPI: int32 __nv_sad(int32 x, int32 y, int32 z);
    """


def saturatef(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_saturatef.html

    CAPI: float32 __nv_saturatef(float32 x);
    """


def scalbn(x: float64, y: int32) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_scalbn.html

    CAPI: float64 __nv_scalbn(float64 x, int32 y);
    """


def scalbnf(x: float32, y: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_scalbnf.html

    CAPI: float32 __nv_scalbnf(float32 x, int32 y);
    """


def signbitd(x: float64) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_signbitd.html

    CAPI: int32 __nv_signbitd(float64 x);
    """


def signbitf(x: float32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_signbitf.html

    CAPI: int32 __nv_signbitf(float32 x);
    """


def sin(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sin.html

    CAPI: float64 __nv_sin(float64 x);
    """


def sincos(x: float64) -> UniTuple(float64, 2):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincos.html

    CAPI: void __nv_sincos(float64 x, float64* sptr, float64* cptr);
    """


def sincosf(x: float32) -> UniTuple(float32, 2):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincosf.html

    CAPI: void __nv_sincosf(float32 x, float32* sptr, float32* cptr);
    """


def sincospi(x: float64) -> UniTuple(float64, 2):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincospi.html

    CAPI: void __nv_sincospi(float64 x, float64* sptr, float64* cptr);
    """


def sincospif(x: float32) -> UniTuple(float32, 2):
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sincospif.html

    CAPI: void __nv_sincospif(float32 x, float32* sptr, float32* cptr);
    """


def sinf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinf.html

    CAPI: float32 __nv_sinf(float32 x);
    """


def sinh(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinh.html

    CAPI: float64 __nv_sinh(float64 x);
    """


def sinhf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinhf.html

    CAPI: float32 __nv_sinhf(float32 x);
    """


def sinpi(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinpi.html

    CAPI: float64 __nv_sinpi(float64 x);
    """


def sinpif(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sinpif.html

    CAPI: float32 __nv_sinpif(float32 x);
    """


def sqrt(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sqrt.html

    CAPI: float64 __nv_sqrt(float64 x);
    """


def sqrtf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_sqrtf.html

    CAPI: float32 __nv_sqrtf(float32 x);
    """


def tan(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tan.html

    CAPI: float64 __nv_tan(float64 x);
    """


def tanf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tanf.html

    CAPI: float32 __nv_tanf(float32 x);
    """


def tanh(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tanh.html

    CAPI: float64 __nv_tanh(float64 x);
    """


def tanhf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tanhf.html

    CAPI: float32 __nv_tanhf(float32 x);
    """


def tgamma(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tgamma.html

    CAPI: float64 __nv_tgamma(float64 x);
    """


def tgammaf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_tgammaf.html

    CAPI: float32 __nv_tgammaf(float32 x);
    """


def trunc(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_trunc.html

    CAPI: float64 __nv_trunc(float64 x);
    """


def truncf(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_truncf.html

    CAPI: float32 __nv_truncf(float32 x);
    """


def uhadd(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uhadd.html

    CAPI: int32 __nv_uhadd(int32 x, int32 y);
    """


def uint2double_rn(i: int32) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2double_rn.html

    CAPI: float64 __nv_uint2double_rn(int32 i);
    """


def uint2float_rd(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_rd.html

    CAPI: float32 __nv_uint2float_rd(int32 in);
    """


def uint2float_rn(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_rn.html

    CAPI: float32 __nv_uint2float_rn(int32 in);
    """


def uint2float_ru(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_ru.html

    CAPI: float32 __nv_uint2float_ru(int32 in);
    """


def uint2float_rz(in_: int32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_uint2float_rz.html

    CAPI: float32 __nv_uint2float_rz(int32 in);
    """


def ull2double_rd(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_rd.html

    CAPI: float64 __nv_ull2double_rd(int64 l);
    """


def ull2double_rn(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_rn.html

    CAPI: float64 __nv_ull2double_rn(int64 l);
    """


def ull2double_ru(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_ru.html

    CAPI: float64 __nv_ull2double_ru(int64 l);
    """


def ull2double_rz(l: int64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2double_rz.html

    CAPI: float64 __nv_ull2double_rz(int64 l);
    """


def ull2float_rd(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_rd.html

    CAPI: float32 __nv_ull2float_rd(int64 l);
    """


def ull2float_rn(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_rn.html

    CAPI: float32 __nv_ull2float_rn(int64 l);
    """


def ull2float_ru(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_ru.html

    CAPI: float32 __nv_ull2float_ru(int64 l);
    """


def ull2float_rz(l: int64) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ull2float_rz.html

    CAPI: float32 __nv_ull2float_rz(int64 l);
    """


def ullmax(x: int64, y: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ullmax.html

    CAPI: int64 __nv_ullmax(int64 x, int64 y);
    """


def ullmin(x: int64, y: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ullmin.html

    CAPI: int64 __nv_ullmin(int64 x, int64 y);
    """


def umax(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umax.html

    CAPI: int32 __nv_umax(int32 x, int32 y);
    """


def umin(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umin.html

    CAPI: int32 __nv_umin(int32 x, int32 y);
    """


def umul24(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umul24.html

    CAPI: int32 __nv_umul24(int32 x, int32 y);
    """


def umul64hi(x: int64, y: int64) -> int64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umul64hi.html

    CAPI: int64 __nv_umul64hi(int64 x, int64 y);
    """


def umulhi(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_umulhi.html

    CAPI: int32 __nv_umulhi(int32 x, int32 y);
    """


def urhadd(x: int32, y: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_urhadd.html

    CAPI: int32 __nv_urhadd(int32 x, int32 y);
    """


def usad(x: int32, y: int32, z: int32) -> int32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_usad.html

    CAPI: int32 __nv_usad(int32 x, int32 y, int32 z);
    """


def y0(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y0.html

    CAPI: float64 __nv_y0(float64 x);
    """


def y0f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y0f.html

    CAPI: float32 __nv_y0f(float32 x);
    """


def y1(x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y1.html

    CAPI: float64 __nv_y1(float64 x);
    """


def y1f(x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_y1f.html

    CAPI: float32 __nv_y1f(float32 x);
    """


def yn(n: int32, x: float64) -> float64:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_yn.html

    CAPI: float64 __nv_yn(int32 n, float64 x);
    """


def ynf(n: int32, x: float32) -> float32:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_ynf.html

    CAPI: float32 __nv_ynf(int32 n, float32 x);
    """



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Common struct and union definitions for tcgen05 (UTCMMA) operations.

This module provides canonical definitions of structs and unions used for
tcgen05 instructions, including shared memory descriptors and instruction
descriptors. These definitions match the hardware specification and can be
reused across different kernels.
"""

import ctypes  # Import at module level for use in device functions
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.cuda.experimental.struct import struct
from numba_cuda_mlir.cuda.experimental.union import union

# ============================================================================
# Instruction Descriptor
# ============================================================================
# Union for tcgen05 instruction descriptor
# Matches the hardware specification for UTCMMA instruction encoding

InstrDescriptorBitfields = struct(
    [
        ("sparse_id2_", types.uint32, 2),
        ("sparse_flag_", types.uint32, 1),
        ("saturate_", types.uint32, 1),
        ("c_format_", types.uint32, 2),
        ("sparse_format_", types.uint32, 1),
        ("a_format_", types.uint32, 3),
        ("b_format_", types.uint32, 3),
        ("a_negate_", types.uint32, 1),
        ("b_negate_", types.uint32, 1),
        ("a_major_", types.uint32, 1),
        ("b_major_", types.uint32, 1),
        ("n_dim_", types.uint32, 6),
        (None, types.uint32, 1),
        ("m_dim_", types.uint32, 5),
        (None, types.uint32, 1),
        ("max_shift_", types.uint32, 2),
    ],
    name="InstrDescriptorBitfields",
)

InstrDescriptor = union(
    [
        ("desc_", types.uint32),
        ("bitfields", InstrDescriptorBitfields),
    ],
    name="InstrDescriptor",
)


# ============================================================================
# Shared Memory Descriptor
# ============================================================================
# Union for tcgen05 shared memory descriptor
# Used to describe shared memory layout for TMA loads

SmemDescriptorBitfields = struct(
    [
        ("start_address_", types.uint64, 14),
        (None, types.uint64, 2),
        ("leading_byte_offset_", types.uint64, 14),
        (None, types.uint64, 2),
        ("stride_byte_offset_", types.uint64, 14),
        ("version_", types.uint64, 2),
        (None, types.uint64, 1),
        ("base_offset_", types.uint64, 3),
        (None, types.uint64, 4),
        (None, types.uint64, 5),
        ("layout_type_", types.uint64, 3),
    ],
    name="SmemDescriptorBitfields",
)

SmemDescriptor = union(
    [
        ("desc_", types.uint64),
        ("reg32_", types.UniTuple(types.uint32, 2)),
        ("reg16_", types.UniTuple(types.uint16, 4)),
        ("bitfields", SmemDescriptorBitfields),
    ],
    name="SmemDescriptor",
)


# ============================================================================
# Device Functions for Building Descriptors
# ============================================================================
# These functions can be called from within JIT-compiled kernels
# They are compiled as device functions using cuda.jit(device=True)


# Import cuda module - use lazy import to avoid circular dependencies
def _get_cuda():
    """Get cuda module, importing lazily to avoid circular imports."""
    import numba_cuda_mlir.cuda as cuda

    return cuda


# Public API functions with default parameters
# These will be compiled as device functions using cuda.jit(device=True)
def tcgen05_build_smem_descriptor(
    smem_ptr,
    leading_byte_offset=16,
    stride_byte_offset=1024,
    base_offset=0,
    layout=2,
):
    """
    Build a tcgen05 shared memory descriptor.

    This function can be called from within JIT-compiled kernels.

    Parameters:
    - smem_ptr: Shared memory pointer (from types.ptr(smem_array))
    - leading_byte_offset: Leading byte offset (default: 16)
    - stride_byte_offset: Stride byte offset (default: 1024)
    - base_offset: Base offset (default: 0)
    - layout: Layout type (default: 2 for SWIZZLE_128B)

    Returns:
    - SmemDescriptor instance with desc_ field ready for use
    """
    smem_desc = SmemDescriptor()
    bitfields = smem_desc.bitfields
    bitfields.start_address_ = ctypes.cast(smem_ptr, ctypes.c_uint32) >> 4
    bitfields.leading_byte_offset_ = leading_byte_offset >> 4
    bitfields.stride_byte_offset_ = stride_byte_offset >> 4
    bitfields.base_offset_ = base_offset
    bitfields.layout_type_ = layout
    smem_desc.bitfields = bitfields
    return smem_desc


def tcgen05_build_idescriptor(
    c_format=1,
    n_dim=16,
    m_dim=8,
):
    """
    Build a tcgen05 instruction descriptor.

    This function can be called from within JIT-compiled kernels.

    Parameters:
    - c_format: C format (default: 1 for F32)
    - n_dim: N dimension (default: 16)
    - m_dim: M dimension (default: 8)

    Returns:
    - InstrDescriptor instance with desc_ field ready for use
    """
    idesc = InstrDescriptor()
    bitfields = idesc.bitfields
    bitfields.c_format_ = c_format
    bitfields.n_dim_ = n_dim
    bitfields.m_dim_ = m_dim
    idesc.bitfields = bitfields
    return idesc


def tcgen05_update_smem_start_address(
    smem_desc,
    offset,
):
    """
    Update the start address of a shared memory descriptor.

    This function can be called from within JIT-compiled kernels.

    Parameters:
    - smem_desc: Existing SmemDescriptor instance
    - offset: Offset to add to the start address (in units of 16 bytes)

    Returns:
    - New SmemDescriptor instance with updated start_address_
    """
    new_desc = SmemDescriptor()
    bitfields = smem_desc.bitfields
    new_bitfields = new_desc.bitfields
    # Copy all fields from the original descriptor
    new_bitfields.start_address_ = bitfields.start_address_ + offset
    new_bitfields.leading_byte_offset_ = bitfields.leading_byte_offset_
    new_bitfields.stride_byte_offset_ = bitfields.stride_byte_offset_
    new_bitfields.version_ = bitfields.version_
    new_bitfields.base_offset_ = bitfields.base_offset_
    new_bitfields.layout_type_ = bitfields.layout_type_
    new_desc.bitfields = new_bitfields
    return new_desc


# Compile as device functions after module is loaded
# This avoids circular import issues by deferring compilation
def _init_device_functions():
    """Initialize device functions by compiling them."""
    cuda = _get_cuda()

    global \
        tcgen05_build_smem_descriptor, \
        tcgen05_build_idescriptor, \
        tcgen05_update_smem_start_address

    # Compile the wrapper functions (with defaults) directly as device functions
    # This matches the pattern used in test examples
    tcgen05_build_smem_descriptor = cuda.jit(device=True, inline="always")(
        tcgen05_build_smem_descriptor
    )
    tcgen05_build_idescriptor = cuda.jit(device=True, inline="always")(tcgen05_build_idescriptor)
    tcgen05_update_smem_start_address = cuda.jit(device=True, inline="always")(
        tcgen05_update_smem_start_address
    )



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Implement slices and various slice computations.
"""

from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda import cgutils
from numba_cuda_mlir.numba_cuda.core.imputils import impl_ret_untracked, Registry

registry = Registry("slicing")
lower = registry.lower
lower_getattr = registry.lower_getattr
lower_cast = registry.lower_cast
lower_constant = registry.lower_constant


# fix_index / fix_slice / get_slice_length build llvmlite slice-arithmetic IR and
# are only called from filtered-out vendored @lower codegen (np/arrayobj,
# cpython/unicode). They are dead on the MLIR path (slicing is handled by
# numba_cuda_mlir.lowering), so their bodies are tombstones.
def fix_index(builder, idx, size):
    raise NotImplementedError("slicing codegen (fix_index) is not used on the MLIR path")


def fix_slice(builder, slice, size):
    raise NotImplementedError("slicing codegen (fix_slice) is not used on the MLIR path")


def get_slice_length(builder, slicestruct):
    raise NotImplementedError("slicing codegen (get_slice_length) is not used on the MLIR path")


def get_slice_bounds(builder, slicestruct):
    """
    Return the [lower, upper) indexing bounds of a slice.
    """
    start = slicestruct.start
    stop = slicestruct.stop
    zero = start.type(0)
    one = start.type(1)
    # This is a bit pessimal, e.g. it will return [1, 5) instead
    # of [1, 4) for `1:5:2`
    is_step_negative = builder.icmp_signed("<", slicestruct.step, zero)
    lower = builder.select(is_step_negative, builder.add(stop, one), start)
    upper = builder.select(is_step_negative, builder.add(start, one), stop)
    return lower, upper


def fix_stride(builder, slice, stride):
    """
    Fix the given stride for the slice's step.
    """
    return builder.mul(slice.step, stride)


def guard_invalid_slice(context, builder, typ, slicestruct):
    """
    Guard against *slicestruct* having a zero step (and raise ValueError).
    """
    if typ.has_step:
        cgutils.guard_null(
            context,
            builder,
            slicestruct.step,
            (ValueError, "slice step cannot be zero"),
        )


def get_defaults(context):
    """
    Get the default values for a slice's members:
    (start for positive step, start for negative step,
     stop for positive step, stop for negative step, step)
    """
    maxint = (1 << (context.address_size - 1)) - 1
    return (0, maxint, maxint, -maxint - 1, 1)


# ---------------------------------------------------------------------------
# The slice structure


@lower(slice, types.VarArg(types.Any))
def slice_constructor_impl(context, builder, sig, args):
    (
        default_start_pos,
        default_start_neg,
        default_stop_pos,
        default_stop_neg,
        default_step,
    ) = [context.get_constant(types.intp, x) for x in get_defaults(context)]

    slice_args = [None] * 3

    # Fetch non-None arguments
    if len(args) == 1 and sig.args[0] is not types.none:
        slice_args[1] = args[0]
    else:
        for i, (ty, val) in enumerate(zip(sig.args, args)):
            if ty is not types.none:
                slice_args[i] = val

    # Fill omitted arguments
    def get_arg_value(i, default):
        val = slice_args[i]
        if val is None:
            return default
        else:
            return val

    step = get_arg_value(2, default_step)
    is_step_negative = builder.icmp_signed("<", step, context.get_constant(types.intp, 0))
    default_stop = builder.select(is_step_negative, default_stop_neg, default_stop_pos)
    default_start = builder.select(is_step_negative, default_start_neg, default_start_pos)
    stop = get_arg_value(1, default_stop)
    start = get_arg_value(0, default_start)

    ty = sig.return_type
    sli = context.make_helper(builder, sig.return_type)
    sli.start = start
    sli.stop = stop
    sli.step = step

    res = sli._getvalue()
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_getattr(types.SliceType, "start")
def slice_start_impl(context, builder, typ, value):
    sli = context.make_helper(builder, typ, value)
    return sli.start


@lower_getattr(types.SliceType, "stop")
def slice_stop_impl(context, builder, typ, value):
    sli = context.make_helper(builder, typ, value)
    return sli.stop


@lower_getattr(types.SliceType, "step")
def slice_step_impl(context, builder, typ, value):
    if typ.has_step:
        sli = context.make_helper(builder, typ, value)
        return sli.step
    else:
        return context.get_constant(types.intp, 1)


@lower("slice.indices", types.SliceType, types.Integer)
def slice_indices(context, builder, sig, args):
    length = args[1]
    sli = context.make_helper(builder, sig.args[0], args[0])

    with builder.if_then(cgutils.is_neg_int(builder, length), likely=False):
        context.fndesc.call_conv.return_user_exc(
            builder,
            ValueError,
            ("length should not be negative",),
        )
    with builder.if_then(cgutils.is_scalar_zero(builder, sli.step), likely=False):
        context.fndesc.call_conv.return_user_exc(
            builder,
            ValueError,
            ("slice step cannot be zero",),
        )

    fix_slice(builder, sli, length)

    return context.make_tuple(builder, sig.return_type, (sli.start, sli.stop, sli.step))


def make_slice_from_constant(context, builder, ty, pyval):
    sli = context.make_helper(builder, ty)
    lty = context.get_value_type(types.intp)

    (
        default_start_pos,
        default_start_neg,
        default_stop_pos,
        default_stop_neg,
        default_step,
    ) = [context.get_constant(types.intp, x) for x in get_defaults(context)]

    step = pyval.step
    if step is None:
        step_is_neg = False
        step = default_step
    else:
        step_is_neg = step < 0
        step = lty(step)

    start = pyval.start
    if start is None:
        if step_is_neg:
            start = default_start_neg
        else:
            start = default_start_pos
    else:
        start = lty(start)

    stop = pyval.stop
    if stop is None:
        if step_is_neg:
            stop = default_stop_neg
        else:
            stop = default_stop_pos
    else:
        stop = lty(stop)

    sli.start = start
    sli.stop = stop
    sli.step = step

    return sli._getvalue()


@lower_constant(types.SliceType)
def constant_slice(context, builder, ty, pyval):
    if isinstance(ty, types.Literal):
        typ = ty.literal_type
    else:
        typ = ty

    return make_slice_from_constant(context, builder, typ, pyval)


@lower_cast(types.misc.SliceLiteral, types.SliceType)
def cast_from_literal(context, builder, fromty, toty, val):
    return make_slice_from_constant(
        context,
        builder,
        toty,
        fromty.literal_value,
    )
